import numpy as np

# Package to do OS-CFAR + Useful filters
__version__ = "1.1.24"

from . import cfar
from . import filters
from . import utils
from . import gaussian_fit
from . import cluster

# OS-CFAR versions
os_cfar = cfar.os_cfar_1d
vwindow_os_cfar = cfar.variable_window_os_cfar_indices

# Filters
baseline_filter = filters.remove_baseline_peaks
lowpass_filter = filters.lowpass_filter
highpass_filter = filters.highpass_filter

median_smoothing = filters.median_filter
mean_smoothing = filters.moving_average_filter
lowess_smoothing = filters.lowess_filter

group_peaks = filters.find_representative_peaks
force_min_dist = filters.enforce_min_distance
check_snr = filters.verify_peaks_snr
force_max_extent = filters.filter_peaks_by_extent_1d

# Utilities
peaks = utils.Peaks
waterfall_axes = utils.WaterFallAxes
waterfall_grid = utils.WaterFallGrid
npz_reader = utils.NpzReader
npz_writer = utils.NpzWriter

# Other values for utilities
reader = utils.NpzReader
writer = utils.NpzWriter

# Gaussian functions
multi_gaussian = gaussian_fit.sum_of_gaussians
multi_scattered_gaussian = gaussian_fit.sum_of_scattered_gaussians
grid_search_gaussian = gaussian_fit.find_best_multi_gaussian_fit

# Other utilities
best_params = {
    "guard_cells": 1,
    "train_cells": 8,
    "rank_k": 0.75,
    "threshold_factor": 0.9,
    "averaging": 2,
    "min_snr": 2,
    "baseline": 0.15,
    "smoothing": "lowess",
}


def do_os_cfar(
    data: np.ndarray,
    guard_cells,
    train_cells,
    rank_k,
    threshold_factor,
    averaging,
    min_dist,
    min_snr,
    baseline,
    smoothing="mean",
    lowess_frac=0.01,
    clustering=False,
    clustering_cutoff=0.7,
):
    """
    Perform OS-CFAR detection and post-processing on 2D array data.

    This function applies Ordered Statistics CFAR (OS-CFAR) detection to a 2D input array
    (e.g., a spectrogram), typically by summing along one axis to produce a 1D time series.
    It then applies smoothing, peak detection, SNR filtering, baseline removal, and optional clustering.

    Args:
        data (np.ndarray): 2D input array (e.g., frequency x time), assumed to be linear power.
        guard_cells (int): Number of guard cells on each side of the cell under test (CUT).
        train_cells (int): Number of training cells on each side of the CUT.
        rank_k (float): Fraction (0-1) of training cells to use for the OS-CFAR rank (e.g., 0.75 for 75th percentile).
        threshold_factor (float): Scaling factor (alpha) for the OS-CFAR threshold.
        averaging (int): Window size for initial smoothing (mean or median).
        min_dist (int): Minimum distance (in samples) between detected peaks.
        min_snr (float): Minimum SNR required for a peak to be kept.
        baseline (float): Threshold factor for removing peaks close to the baseline.
        smoothing (str, optional): Smoothing method for the time series ('mean', 'median', 'lowess', or None). Default is 'mean'.
        lowess_frac (float, optional): Fraction of data for LOWESS smoothing if used. Default is 0.01.
        clustering (int or bool, optional): If int > 0, applies DBSCAN clustering to peaks with this as min_samples. Default is False.
        clustering_cutoff (float, optional): Cutoff value for clustering. Default is 0.7.

    Returns:
        Peaks: Peaks object containing detected peak indices and the OS-CFAR threshold array.

    Raises:
        TypeError: If input data is not a 2D NumPy array.
    """

    if len(data.shape) != 2:
        raise TypeError(
            f"Data should be two dimensional only! Received {len(data.shape)} dimensions."
        )

    x = np.arange(data.shape[1])
    ts = np.sum(data, 0)
    stdev = np.std(data, 0)
    mts = np.mean(data, 0)

    if smoothing == "mean":
        filtered = mean_smoothing(ts, averaging)
    elif smoothing == "median":
        filtered = median_smoothing(ts, averaging)
    elif smoothing == "lowess":
        filtered = lowess_smoothing(x, ts, lowess_frac)
    else:
        filtered = ts

    res = os_cfar(
        filtered,
        guard_cells,
        train_cells,
        int(rank_k * 2 * train_cells),
        threshold_factor,
    )

    pk = res[0]
    pk = force_min_dist(list(pk), filtered, min_dist)
    pk = check_snr(filtered, list(pk), min_dist, min_snr)
    pk = group_peaks(filtered, list(pk), min_dist)
    pk = baseline_filter(mts, pk, stdev, baseline)

    if type(clustering) == int and clustering != 0:
        pk = cluster.cluster_peaks_ph(pk, filtered[pk], clustering, clustering_cutoff)

    return peaks((pk, res[1]))



#!/usr/bin/env python3

from __future__ import annotations

import argparse
import re
from pathlib import Path


MODULE_ORDER = [
    "oscfar",
    "oscfar-cfar",
    "oscfar-cluster",
    "oscfar-filters",
    "oscfar-gaussian_fit",
    "oscfar-reload",
    "oscfar-setup",
    "oscfar-utils",
]

DOC_TARGETS = {
    "oscfar": {"file": "oscfar.md", "title": "General functions", "nav_order": 1},
    "oscfar-cfar": {
        "file": "oscfar-cfar.md",
        "title": "CFAR Algorithms",
        "nav_order": 2,
    },
    "oscfar-filters": {
        "file": "oscfar-filters.md",
        "title": "Filtering",
        "nav_order": 3,
    },
    "oscfar-cluster": {
        "file": "oscfar-cluster.md",
        "title": "Clustering",
        "nav_order": 4,
    },
    "oscfar-gaussian_fit": {
        "file": "oscfar-gaussian.md",
        "title": "Fitting",
        "nav_order": 5,
    },
    "oscfar-utils": {"file": "oscfar-utils.md", "title": "Utilities", "nav_order": 6},
}


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Split python_docstring_markdown output into docs pages."
    )
    parser.add_argument(
        "input_file", type=Path, help="Path to generated markdown (temp.txt)"
    )
    parser.add_argument(
        "--docs-dir",
        type=Path,
        default=Path("docs"),
        help="Directory where docs markdown files are written",
    )
    return parser.parse_args()


def find_anchor_block(
    content: str, anchor: str, anchor_positions: dict[str, int]
) -> str:
    start = anchor_positions.get(anchor)
    if start is None:
        raise ValueError(f"Anchor '{anchor}' was not found in generated documentation.")

    next_pos = len(content)
    anchor_index = MODULE_ORDER.index(anchor)
    for next_anchor in MODULE_ORDER[anchor_index + 1 :]:
        candidate = anchor_positions.get(next_anchor)
        if candidate is not None and candidate > start:
            next_pos = candidate
            break

    return content[start:next_pos].strip() + "\n"


def normalize_module_block(block: str) -> str:
    lines = block.splitlines()

    module_heading_idx = next(
        (i for i, line in enumerate(lines) if line.startswith("## 🅼 ")), None
    )
    functions_heading_idx = next(
        (i for i, line in enumerate(lines) if line.strip() == "### Functions"), None
    )

    if module_heading_idx is not None:
        if (
            module_heading_idx + 1 >= len(lines)
            or lines[module_heading_idx + 1].strip() != "{: .no_toc }"
        ):
            lines.insert(module_heading_idx + 1, "{: .no_toc }")
            lines.insert(module_heading_idx + 2, "")
            if (
                functions_heading_idx is not None
                and functions_heading_idx > module_heading_idx
            ):
                functions_heading_idx += 2

    if module_heading_idx is not None and functions_heading_idx is not None:
        del lines[module_heading_idx + 1 : functions_heading_idx]
        lines.insert(module_heading_idx + 1, "")
        functions_heading_idx = module_heading_idx + 2

    if functions_heading_idx is not None:
        replacement = [
            "### Functions",
            "{: .no_toc .text-delta }",
            "",
            "1. TOC",
            "{:toc}",
            "",
        ]
        lines[functions_heading_idx : functions_heading_idx + 1] = replacement

    cleaned = "\n".join(lines).strip() + "\n"
    cleaned = re.sub(r"\n{3,}", "\n\n", cleaned)
    return cleaned


def get_existing_front_matter(path: Path) -> str | None:
    if not path.exists():
        return None

    text = path.read_text(encoding="utf-8")
    if not text.startswith("---\n"):
        return None

    end = text.find("\n---\n", 4)
    if end == -1:
        return None

    return text[: end + 5].strip() + "\n"


def default_front_matter(title: str, nav_order: int) -> str:
    return (
        "---\n"
        f"title: {title}\n"
        "layout: page\n"
        "parent: Documentation\n"
        f"nav_order: {nav_order}\n"
        "---\n"
    )


def write_module_docs(content: str, docs_dir: Path) -> None:
    anchor_positions = {
        anchor: content.find(f'<a name="{anchor}"></a>') for anchor in MODULE_ORDER
    }
    anchor_positions = {k: v for k, v in anchor_positions.items() if v != -1}

    for anchor, target in DOC_TARGETS.items():
        block = find_anchor_block(content, anchor, anchor_positions)
        normalized = normalize_module_block(block)

        output_path = docs_dir / target["file"]
        front_matter = get_existing_front_matter(output_path)
        if front_matter is None:
            front_matter = default_front_matter(
                title=target["title"], nav_order=target["nav_order"]
            )

        final_text = f"{front_matter}\n{normalized}"
        output_path.write_text(final_text, encoding="utf-8")
        print(f"Updated {output_path}")


def main() -> None:
    args = parse_args()
    input_path = args.input_file
    docs_dir = args.docs_dir

    if not input_path.exists():
        raise FileNotFoundError(f"Input file not found: {input_path}")

    docs_dir.mkdir(parents=True, exist_ok=True)

    content = input_path.read_text(encoding="utf-8")
    write_module_docs(content, docs_dir)


if __name__ == "__main__":
    main()
